/********************************************************************************
* Advanced Database Topics
* Dave Valentine
* Indexing
*
* Index Demo Scripts
********************************************************************************/

USE AdventureWorks2012;
GO

SET STATISTICS IO ON;

/***********************************************************************************
* Create Production.BackupTransactionHistory using the INTO statement.  Adding the 
* primary key constraint to the table as well.
***********************************************************************************/
-- Remove Production.BackupTransactionHistory if exists
IF OBJECT_ID('Production.BackupTransactionHistory') IS NOT NULL 
   DROP TABLE Production.BackupTransactionHistory;
GO


SELECT   TransactionID
       , ProductID
       , ReferenceOrderID
       , ReferenceOrderLineID
       , TransactionDate
       , TransactionType
       , Quantity
       , ActualCost
       , ModifiedDate
INTO     Production.BackupTransactionHistory
FROM     Production.TransactionHistory; 
GO

/***********************************************************************************
* Running the following query will result in an Table scan.  Using Ctrl-M
* will display the Execution Plan that shows the Table Scan. 
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID = 843;
GO






/***********************************************************************************
*  This script will show that we have a HEAP because the index type = 0.
   
   We will convert this heap to a clustered index with a primary key
***********************************************************************************/
IF EXISTS (
	SELECT 1 FROM sys.tables t
	INNER JOIN sys.indexes i on i.object_id = t.object_id
	WHERE i.type = 0 AND t.name = 'BackupTransactionHistory'
) SELECT 'We have a HEAP'

-- Create the PRIMARY KEY
ALTER TABLE Production.BackupTransactionHistory 
ADD CONSTRAINT PK_BackupTransactionHistory PRIMARY KEY (TransactionID) 
GO


/***********************************************************************************
*  This script will show that we have are scanning a clustlered index
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID = 843;
GO




/***********************************************************************************
*  Part I - Sorting Issues
*
*  Before we add addtional indexes to fix the index scan lets take a look at a 
*  problem with sorting.
*  
*  This script will return a table scan as well, but will return all the records 
*  in the table
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
GO




/***********************************************************************************
*  This script will return all sorts of new operators.  We have still have our 
*  Clusltered Index Scan, however we have introducted a Sort and Parallelism operator.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
ORDER BY ActualCost;






/***********************************************************************************
*  This script will should return the same operators, but look at what has happened
*  with the estimated subtree cost.
***********************************************************************************/
SELECT   *
FROM     Production.BackupTransactionHistory
ORDER BY ActualCost;










/***********************************************************************************
* PART II - Covered Indexes
*
* Running the following query will result in an index scan.  Using Ctrl-M
* will display the Execution Plan that shows the Index Scan. 
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID = 843;
GO


/***********************************************************************************
* By adding a NONCLUSTERED INDEX on ProductID to Production.BackupTransactionHistory 
* we can eliminate the index scan
***********************************************************************************/
CREATE NONCLUSTERED INDEX IX_BackupTransactionHistory_ProductID 
ON Production.BackupTransactionHistory(ProductID ASC);
GO



/***********************************************************************************
* Running the query again will result in an index seek.  However, we now have
* a key lookup.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID = 843;
GO



/***********************************************************************************
* Adding a covering index will elminate the key lookup.  Start by removing the old
* index, then we can create the covering index.
***********************************************************************************/
--Drop the existing index on product id
DROP INDEX IX_BackupTransactionHistory_ProductID 
ON Production.BackupTransactionHistory;
GO


-- Create covering NONCLUSTERED INDEX on ProductID INCLUDE SELECT columns.
CREATE NONCLUSTERED INDEX IX_BackupTransactionHistory_ProductID 
ON Production.BackupTransactionHistory(ProductID ASC)
INCLUDE(TransactionType, Quantity, ActualCost);
GO



/***********************************************************************************
* Running the query again will result in an index seek without the key lookup.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID = 843;
GO








/***********************************************************************************
* So Index Seeks are perfect????
*
* Take a look at these two queries.  Are index seeks going to solve all your problems?
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ProductID > 0;

SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory;
GO






/***********************************************************************************
* PART III - Filtered Indexes
*
* Running the following query for a range of values greater than $2400.00 again
* yields an index scan and page reads.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ActualCost > 2400;
GO



/***********************************************************************************
* Adding a covering NONCLUSTERED INDEX on ActualCost to 
* Production.BackupTransactionHistory will produce an index seek and reduce the 
* logical page reads.
***********************************************************************************/
CREATE NONCLUSTERED INDEX IX_BackupTransactionHistory_ActualCost 
ON Production.BackupTransactionHistory(ActualCost ASC)
INCLUDE(TransactionType, Quantity);
GO



/***********************************************************************************
* Running the query again shows that the covering NONCLUSTERED INDEX is being 
* used, and the page reads are significantly less.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ActualCost > 2400;
GO



/***********************************************************************************
* Adding a filtered NONCLUSTERED INDEX on ActualCost to 
* Production.BackupTransactionHistory will reduces the index size and the 
* logical page reads.
***********************************************************************************/
CREATE NONCLUSTERED INDEX IX_BackupTransactionHistory_Filtered 
ON Production.BackupTransactionHistory(ActualCost ASC)
INCLUDE(TransactionType, Quantity)
WHERE ActualCost > 2000;
GO



/***********************************************************************************
* Running the query again shows that the filtered NONCLUSTERED INDEX is being 
* used, and the page reads are less.
***********************************************************************************/
SELECT   TransactionType
       , Quantity
       , ActualCost
FROM     Production.BackupTransactionHistory
WHERE    ActualCost > 2400;
GO



/***********************************************************************************
* This Script shows the index type, the key columns used by an index, the included
* columns for the index, and the index size in KB and MB.
*
* Running this script shows that we have three indexes on the 
* Production.BackupTransactionHistory table.  Notice the size difference between
* the different types of indexes.
***********************************************************************************/
DECLARE @TableName sysname = 'Production.BackupTransactionHistory';
SELECT   TableName = s.name + '.' + o.name
       , IndexName = i.name
       , IndexType = i.type_desc
       , IndexKeyColumns = SUBSTRING((
                                      SELECT ', ' + c.name
                                      FROM   sys.index_columns ic
                                      JOIN   sys.columns c ON c.column_id = ic.column_id
                                                              AND c.object_id = ic.object_id
                                      WHERE  ic.index_id = i.index_id
                                             AND ic.object_id = i.object_id
                                             AND ic.is_included_column = 0
                                      ORDER BY key_ordinal FOR XML PATH('')), 3, 1000)
       , IncludedColumns = SUBSTRING((
                                      SELECT ', ' + c.name
                                      FROM   sys.index_columns ic
                                      JOIN   sys.columns c ON c.column_id = ic.column_id
                                                              AND c.object_id = ic.object_id
                                      WHERE  ic.index_id = i.index_id
                                             AND ic.object_id = i.object_id
                                             AND ic.is_included_column = 1
                                      ORDER BY c.name FOR XML PATH('')), 3, 1000)
       , IndexSizeKB = CAST(ps.used_page_count * 8 AS int)
	   , IndexSizeMB = CAST(ps.used_page_count * 8 / 1024.0 AS decimal(9,2))	   
FROM     sys.objects o
JOIN     sys.schemas s ON s.schema_id = o.schema_id
JOIN     sys.indexes i ON i.object_id = o.object_id
JOIN     sys.dm_db_partition_stats ps ON ps.index_id = i.index_id
                                         AND ps.object_id = i.object_id
                                         AND ps.object_id = o.object_id
WHERE    o.object_id = OBJECT_ID(@TableName);
GO


/********************************************************************************
* END Index Demo Scripts
********************************************************************************/
