/********************************************************************************
* Advanced Database Topics
* Dave Valentine
* Indexing
*
* PART II  Indexing XML Demo Scripts
********************************************************************************/

USE AdventureWorks2012;
GO

/***********************************************************************************
* Lets start by taking a look at the XML
***********************************************************************************/
SELECT top 5 * FROM Sales.Store

/***********************************************************************************
* Cleanup the existing INDEX
***********************************************************************************/
DROP INDEX PXML_Store_Demographics ON Sales.Store ;


/***********************************************************************************
* Lets take a look at a couple XML based queries to see what the execution plans look like
***********************************************************************************/
WITH XMLNAMESPACES ('http://schemas.microsoft.com/sqlserver/2004/07/adventure-works/StoreSurvey' AS ssns)
SELECT BusinessEntityID, Name, SalesPersonID
FROM Sales.Store
WHERE Demographics.exist('(/ssns:StoreSurvey/ssns:Internet[.="DSL"])')=1;


WITH XMLNAMESPACES ('http://schemas.microsoft.com/sqlserver/2004/07/adventure-works/StoreSurvey' AS ssns)
SELECT Name
	 , AnnualSales = Demographics.value('(/ssns:StoreSurvey/ssns:AnnualSales )[1]','int')
	 , AnnualRevenue = Demographics.value('(/ssns:StoreSurvey/ssns:AnnualRevenue )[1]','int')
	 , Internet = Demographics.value('(/ssns:StoreSurvey/ssns:Internet)[1]','varchar(30)')
FROM Sales.Store
WHERE SalesPersonID = 277

/***********************************************************************************
* Create the primary XML Index, which will index all tags, nodes, and paths.
***********************************************************************************/
CREATE PRIMARY XML INDEX XIX_Sales_Demographics ON Sales.Store (Demographics);


/***********************************************************************************
* Looking at the XML Query again we see the plan has improved, but there is room
* for improvement
***********************************************************************************/
WITH XMLNAMESPACES ('http://schemas.microsoft.com/sqlserver/2004/07/adventure-works/StoreSurvey' AS ssns)
SELECT BusinessEntityID, Name, SalesPersonID
FROM Sales.Store
WHERE Demographics.exist('(/ssns:StoreSurvey/ssns:Internet[.="DSL"])')=1;




/***********************************************************************************
* Lets create the secondary indexes for PATH, VALUE and PROPERTY
***********************************************************************************/
-- PATH Indexes help with checking for the existence of a path
CREATE XML INDEX XIX_PATH_Sales_Demographics ON Sales.Store (Demographics)
  USING XML INDEX XIX_Sales_Demographics FOR PATH;

-- VALUE Indexes help with specific value searching for displaying
CREATE XML INDEX XIX_VALUE_Sales_Demographics ON Sales.Store (Demographics)
  USING XML INDEX XIX_Sales_Demographics FOR VALUE;

-- PROPERTY Indexes help combo queries types of queries, querying both base table data \
--          and the XML document. 
CREATE XML INDEX XIX_PROPERTY_Sales_Demographics ON Sales.Store (Demographics)
  USING XML INDEX XIX_Sales_Demographics FOR PROPERTY;


/***********************************************************************************
* Looking at the XML Query again we see the plan has improved significantly, 
* primarily due to the PATH index
***********************************************************************************/
WITH XMLNAMESPACES ('http://schemas.microsoft.com/sqlserver/2004/07/adventure-works/StoreSurvey' AS ssns)
SELECT BusinessEntityID, Name, SalesPersonID
FROM Sales.Store
WHERE Demographics.exist('(/ssns:StoreSurvey/ssns:Internet[.="DSL"])')=1;

/***********************************************************************************
* Looking at the XML Query again we see the plan has improved, 
* primarily due to the VALUE index
***********************************************************************************/
WITH XMLNAMESPACES ('http://schemas.microsoft.com/sqlserver/2004/07/adventure-works/StoreSurvey' AS ssns)
SELECT Name
	 , AnnualSales = Demographics.value('(/ssns:StoreSurvey/ssns:AnnualSales )[1]','int')
	 , AnnualRevenue = Demographics.value('(/ssns:StoreSurvey/ssns:AnnualRevenue )[1]','int')
	 , Internet = Demographics.value('(/ssns:StoreSurvey/ssns:Internet)[1]','varchar(30)')
FROM Sales.Store
WHERE SalesPersonID = 277;


/********************************************************************************
* END PART II  Indexing XML Demo Scripts
********************************************************************************/






