/********************************************************************************
* Advanced Database Topics
* Dave Valentine
* Partitioning
*
* PART III  Database Partitioning Demo Scripts
********************************************************************************/

USE AdventureWorksDW2012;
GO

/********************************************************************************
* *************** DISCLAIMER ***************
* Database Partitioning can have a negative impact on your query performance.
* If you are implementing partitioning for query performance, you may be 
* Dissatisfied with the results 
* 
*
* Lets start by looking at the data and determining a good partitioning strategy.
* It looks like MovementDate would be a good candidate to partition on.
********************************************************************************/
SELECT YEAR(MovementDate), COUNT(1) FROM FactProductInventory
GROUP BY YEAR(MovementDate)




/********************************************************************************
* PARTITION FUNCTION  
* The first step in database partitioning is to decide on the partition function
* and how the data will be stored in each partition.  Here we are going to use 
* RANGE RIGHT, mainly because we are using dates, and implementing multiple extra
* partitions for swapping in and out later.
********************************************************************************/
CREATE PARTITION FUNCTION ProductInventoryPF (DATE)
AS RANGE RIGHT FOR VALUES (
	'20100101','20110101','20120101','20130101','20140101','20150101','20160101','20170101'
);













/********************************************************************************
* ADD FILEGROUP  
* The next step is to create the file groups on the database.  Notice that we are
* creating one more that what is defined in the partition function?  This is
* needed to capture any values that fall below our threshold, we will see this
* when we defined the partition schema.
********************************************************************************/
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2009
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2010
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2011
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2012
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2013
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2014
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2015
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2016
GO
ALTER DATABASE AdventureWorksDW2012 ADD FILEGROUP ProductInventory2017
GO


















/********************************************************************************
* ADD DATABASE FILES  
* The next step is to create the actual database files and assign to the newly
* created database file groups to the specific database files.
********************************************************************************/
ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2009, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2009.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2009;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2010, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2010.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2010;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2011, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2011.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2011;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2012, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2012.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2012;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2013, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2013.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2013;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2014, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2014.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2014;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2015, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2015.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2015;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2016, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2016.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2016;
GO

ALTER DATABASE AdventureWorksDW2012 ADD FILE (
	NAME=ProductInventory2017, 
	FILENAME = 'C:\Development\Advanced Database Topics\DataFiles\ProductInventory2017.ndf', 
	SIZE=50MB,  FILEGROWTH=10MB
) TO FILEGROUP ProductInventory2017;
GO















/********************************************************************************
* CREATE PARTITION SCHEME  
* The next step is to create the partition scheme, the partition scheme uses the
* partition function to specify which partition data will be stored. The order
* the file groups are specified, dictate where the data is stored.  So using
* meaningful names helps to ensure proper order.  Notice here that we have
* more file groups than values specified in the partition function. 
********************************************************************************/
CREATE PARTITION SCHEME ProductInventoryPS
AS PARTITION ProductInventoryPF
TO (ProductInventory2009, ProductInventory2010, ProductInventory2011
  , ProductInventory2012, ProductInventory2013, ProductInventory2014
  , ProductInventory2015, ProductInventory2016, ProductInventory2017);








/********************************************************************************
* Let take a quick look at the File Group Details to make sure everything is 
* setup correctly
********************************************************************************/
SELECT [pf_name]
      ,[partition_scheme_name]
      ,[partition_number]
      ,[partition_filegroup]
      ,[pf_type_desc]
      ,[pf_fanout]
      ,[boundary_value_on_right]
      ,[object_name]
      ,[range_value]
      ,[num_rows]
      ,[reserved_mb_all_indexes]
      ,[num_indexes]
  FROM [dbo].[FileGroupDetail]








/********************************************************************************
* READY, SET, WAIT
* Now that we have defined the partition function, file groups, added database 
* files, and created our partition schema, we are ready to create the table
* partition.  Before we can apply the partition scheme to our table, we need 
* to remove the existing primary key on the FactProductInventory table.  If this
* was a new table we would not have to take this step.
********************************************************************************/
ALTER TABLE [dbo].[FactProductInventory] DROP CONSTRAINT [PK_FactProductInventory]
GO















/********************************************************************************
* READY, SET, PARTITION
* Now we are finally ready to add the partition scheme to a table.  We do this by 
* creating a new clustered index and specifying the partition schema as the 
* file group.  We need to pass the column name to the partition scheme, as this 
* tells the partition function which file group to store the record into.
********************************************************************************/
ALTER TABLE [dbo].[FactProductInventory] ADD  CONSTRAINT [PK_FactProductInventory] PRIMARY KEY CLUSTERED 
(
	[ProductKey] ASC,
	MovementDate ASC
) ON ProductInventoryPS (MovementDate)
GO












/********************************************************************************
* Let take aanother look at the File Group Details to make sure everything was 
* loaded correctly
********************************************************************************/
SELECT [pf_name]
      ,[partition_scheme_name]
      ,[partition_number]
      ,[partition_filegroup]
      ,[pf_type_desc]
      ,[pf_fanout]
      ,[boundary_value_on_right]
      ,[object_name]
      ,[range_value]
      ,[num_rows]
      ,[reserved_mb_all_indexes]
      ,[num_indexes]
  FROM [dbo].[FileGroupDetail]








/********************************************************************************
* SWITCH IN
* Now here is where the true power of table partitioning comes in, switching
* a table into the table partition.
********************************************************************************/



/********************************************************************************
* Lets start out by creating the table that we are going to move into the 
* table partition.  Things to notice here, the primary key definition on the 
* SwitchFactProductInventory table matches the primary key definition 
* defined on the FactProductInventory table.  Also, the file group we are using
* is ProductInventory2015.
********************************************************************************/
CREATE TABLE [dbo].[SwitchFactProductInventory](
	[ProductKey] [int] NOT NULL,
	[DateKey] [int] NOT NULL,
	[MovementDate] [date] NOT NULL,
	[UnitCost] [money] NOT NULL,
	[UnitsIn] [int] NOT NULL,
	[UnitsOut] [int] NOT NULL,
	[UnitsBalance] [int] NOT NULL,
	CONSTRAINT [PK_SwitchFactProductInventory] PRIMARY KEY CLUSTERED ([ProductKey] ASC,[MovementDate] ASC)
) ON ProductInventory2015
GO








/********************************************************************************
* Lets populate the table with 2015 data.  Typically this is where a ETL work
* and other long running process would happen to scrub and validate data.
********************************************************************************/
INSERT INTO [dbo].[SwitchFactProductInventory] (ProductKey, DateKey, MovementDate, UnitCost, UnitsIn, UnitsOut, UnitsBalance)
SELECT ProductKey
	 , DateKey = DateKey + 10000
	 , MovementDate = DATEADD(YEAR,1,MovementDate)
	 , UnitCost
 	 , UnitsIn 
	 , UnitsOut
	 , UnitsBalance 
FROM FactProductInventory 
WHERE YEAR(MovementDate) = 2014










/********************************************************************************
* It is a requirement that constraints are added to the 
* SwitchFactProductInventory table to ensure that that the values will fit into
* the defined partition range.  The reason for this is the SQL Server engine does 
* not validate the data while switching the table into the table partition.  
* However, it does check to ensure the constraints have been added.
********************************************************************************/
ALTER TABLE [dbo].[SwitchFactProductInventory] WITH CHECK
ADD CONSTRAINT CHK_SwitchFactProductInventory_20150101
CHECK (MovementDate >= '20150101');
GO

ALTER TABLE [dbo].[SwitchFactProductInventory] WITH CHECK
ADD CONSTRAINT CHK_SwitchFactProductInventory_20151231
CHECK (MovementDate < '20151231')
GO












/********************************************************************************
* Lets try to switch SwitchFactProductInventory table into in the table 
* partition.  Using the $PARTITION system function we can call our partition
* to get the proper partition id, or we could take another look at the 
* TablePartition view. 
********************************************************************************/
SELECT $PARTITION.ProductInventoryPF('20151215');
-- OR --
SELECT [pf_name]
      ,[partition_scheme_name]
      ,[partition_number]
      ,[partition_filegroup]
      ,[pf_type_desc]
      ,[pf_fanout]
      ,[boundary_value_on_right]
      ,[object_name]
      ,[range_value]
      ,[num_rows]
      ,[reserved_mb_all_indexes]
      ,[num_indexes]
  FROM [dbo].[FileGroupDetail];







/********************************************************************************
* SWITCH IN
********************************************************************************/
ALTER TABLE [dbo].[SwitchFactProductInventory]
SWITCH TO FactProductInventory PARTITION 7










/********************************************************************************
* Based on the error, there is a foreign key that is missing from the 
* SwitchFactProductInventory table, actually there are two.   Lets get them
* added into table.
********************************************************************************/

ALTER TABLE [dbo].[SwitchFactProductInventory]  WITH CHECK ADD  CONSTRAINT [FK_SwitchFactProductInventory_DimDate] FOREIGN KEY([DateKey])
REFERENCES [dbo].[DimDate] ([DateKey])
GO

ALTER TABLE [dbo].[SwitchFactProductInventory] CHECK CONSTRAINT [FK_SwitchFactProductInventory_DimDate]
GO

ALTER TABLE [dbo].[SwitchFactProductInventory]  WITH CHECK ADD  CONSTRAINT [FK_SwitchFactProductInventory_DimProduct] FOREIGN KEY([ProductKey])
REFERENCES [dbo].[DimProduct] ([ProductKey])
GO

ALTER TABLE [dbo].[SwitchFactProductInventory] CHECK CONSTRAINT [FK_SwitchFactProductInventory_DimProduct]
GO







/********************************************************************************
* SWITCH IN
********************************************************************************/
ALTER TABLE [dbo].[SwitchFactProductInventory]
SWITCH TO FactProductInventory PARTITION 7




















/********************************************************************************
* Let take aanother look at the File Group Details to make sure that our table
* was loaded correctly into the partition
********************************************************************************/
SELECT [pf_name]
      ,[partition_scheme_name]
      ,[partition_number]
      ,[partition_filegroup]
      ,[pf_type_desc]
      ,[pf_fanout]
      ,[boundary_value_on_right]
      ,[object_name]
      ,[range_value]
      ,[num_rows]
      ,[reserved_mb_all_indexes]
      ,[num_indexes]
  FROM [dbo].[FileGroupDetail];




/********************************************************************************
* We can take a quick look at the SwitchFactProductInventory table and see that
* is is empty so it's safe to delete it.
********************************************************************************/

SELECT * FROM [dbo].[SwitchFactProductInventory]


DROP TABLE [dbo].[SwitchFactProductInventory]

/********************************************************************************
* PART III  Database Partitioning Demo Scripts
********************************************************************************/

